import pandas as _pd
import os as _os

def path(factor):
    '''
    factor - str - factor name (use names() to find out all factor names included)
    returns the file name of the factor (e.g path('Total Assets') gives 'balance')
    '''
    # can be a loop or sommething
    # Still in abs path, need changes
    p=_os.path.abspath(_os.path.join(_os.path.dirname(__file__),_os.pardir))
    path=_os.path.join(p,'database')
    
    balance=set(_pd.read_csv(path+'/AAPL/AAPL_balance.csv')['Statements'].tolist())
    if factor in balance:
        return 'balance'
    del balance
    income=set(_pd.read_csv(path+'/AAPL/AAPL_income.csv')['Statements'].tolist())
    if factor in income:
        return 'income'
    del income
    cash_flow=set(_pd.read_csv(path+'/AAPL/AAPL_cash_flow.csv')['Statements'].tolist())
    if factor in cash_flow:
        return 'cash_flow'
    del cash_flow
    trading=set(_pd.read_csv(path+'/AAPL/AAPL_Trading_Information.csv')['0'].tolist())
    if factor in trading:
        return 'Trading_Information'
    del trading
    financial=set(_pd.read_csv(path+'/AAPL/AAPL_Financial_Highlights.csv')['0'].tolist())
    if factor in financial:
        return 'Financial_Hightlights'
    del financial
    valuation=set(_pd.read_csv(path+'/AAPL/AAPL_Valuation_Measures.csv')['0'].tolist())
    if factor in valuation:
        return 'Valuation_Measures'
    del valuation
    summary=set(list(_pd.read_csv(path+'/AAPL/AAPL_Summary.csv')))
    if factor in summary:
        return 'Summary'
    del summary
    raise ValueError("Factor '"+factor+"' not in database")



import os as _os
import gc as _gc
import pandas as _pd
import numpy as _np

p = _os.path.abspath(_os.path.join(_os.path.dirname(__file__), _os.pardir))
global datapath
datapath = _os.path.join(p, 'database')




def database_count():
    '''
    prints out the total number of companies and sheets in database
    '''
    a=len(_os.listdir(datapath))-2
    b=_os.walk(datapath) # generator
    c=[1]
    c=len([c[0]+1 for root, dirs, files in b for name in files])-12
    print("Total number of companies contained: {}".format(a))
    print("Total number of detailed sheets: {}".format(c))
    _gc.collect()


        
def factors_names(sheet=None):
    if sheet is None:
        dfs=(_pd.read_csv(_os.path.join(datapath, 'AAPL', c)) for c in _os.listdir(_os.path.join(datapath, 'AAPL'))) # generator
        r=[df.iloc[0:-1,0].values[1:] for df in dfs]
        r=[i.tolist()[j] for i in r for j in range(len(i.tolist()))]+list(_pd.read_csv(_os.path.join(datapath, 'AAPL', 'AAPL_Summary.csv')))[1:]
        r=_np.array(r, dtype='str')
    else:
        if sheet not in sheets_names():
            raise ValueError('Parameter sheet is wrong, use sheets_names() to find all sheets names')
        file='AAPL_{}.csv'.format(sheet)
        if sheet=='Summary':
            r=_np.array(list(_pd.read_csv(_os.path.join(datapath, 'AAPL', 'AAPL_Summary.csv')))[1:])
        else:
            r=_pd.read_csv(_os.path.join(datapath, 'AAPL', file)).iloc[0:-1, 0].values[1:]
    _gc.collect()
    return r


def sheets_names():
    '''
    returns a list of ALL sheets names (not all companies have all of them)
    '''
    names = ['major_holders', 'top_institutional_holders', 'top_mutual_fund_holders',
             'Trading_Information', 'Financial_Highlights', 'Valuation_Measures',
             'Executives', 'Description',
             'Earnings_Estimate', 'Revenue_Estimate', 'Earnings_History',
             'EPS_Trend', 'EPS_Revisions', 'Growth_Estimates',
             'stats', 'statements', 'reports',
             'Executives', 'Description', 'analysis', 'Summary',
             'balance', 'cash_flow', 'income']
    return names


def code_count():
    '''
    prints out some of the information of the this package
    '''
    
    path=_os.path.abspath('../')
    dirs=_os.listdir(path)
    dirs=[i for i in dirs if 'database' not in i] # get rid of database stuff
    #dirs.remove('__pycache__')
    count={}

    for d in dirs:
        cnt=0
        if '.py' in d:
            f=open(_os.path.join(path, d), mode='r')
            while True: # reduce memory usage
                try:
                    line=f.readline()
                except UnicodeDecodeError:
                    cnt+=1
                    continue
                if not line:
                    break
                else:
                    cnt+=1
            f.close()
        elif _os.path.isdir(_os.path.join(path, d)):
            for dd in _os.listdir(_os.path.join(path, d)):
                if '.py' in dd:
                    f=open(_os.path.join(path, d, dd), mode='r')
                    while True: # reduce memory usage
                        try:
                            line=f.readline()
                        except UnicodeDecodeError:
                            cnt+=1
                            continue
                        if not line:
                            break
                        else:
                            cnt+=1
                    f.close()
        if cnt!=0:
            count[d]=cnt
    return sum(count.values())
    



import numpy as _np
import re


def _decimal(array):  # eliminate ',' in numbers
    d = _np.vectorize(lambda x: float(re.compile(',').sub('', x)))
    return d(array)


def money_digits(array): # not done, probably not useful
    digits = [i[-1] for i in array]
    d = '|'.join([str(i in ['B', 'M', 'K']) for i in digits])
    t = {'B': 1000000000, 'M': 1000000, 'K': 1000}
    if eval(d):
        for i in range(len(array)):
            array[i] = array[i] * t[array[i][-1]]
            pass


def factor(df, factor, year=True):
    # dimensions of f are different for 'factor' and ['factor']
    if isinstance(factor, str):
        if year == True:  # all years
            f = _decimal(df[df['Statements'].isin([factor])].values[0][1:])
        else:  # single factor with specified years
            temp = lambda x, y: [i for j in x for i in y if str(j) in i]  # map 2016 to '9/26/2016'
            years = temp(year, list(df))
            f = _decimal(df.loc[df.Statements == factor, years])
    elif isinstance(factor, list):

        if year == True:
            f = _decimal(_np.array([d[1:] for d in df[df['Statements'].isin(factor)].values]))
        elif year == 'NEWEST':  # multiple factors with the newest year
            f = _decimal(_np.array([[d[1] for d in df[df['Statements'].isin(factor)].values]]))
        else:
            temp = lambda x, y: [i for j in x for i in y if str(j) in i]
            years = temp(year, list(df))
            f = _decimal([df.loc[df.Statements == fac, years].values for fac in factor])
        # big bug here, not handling specific year list, eg. [2018, 2017]
    return f



import pandas as pd
from pyquery import PyQuery as pq

def get_executives(html):
    doc=pq(html)
    text=doc('#Main table').text().split('\n')
    t=[text[i:i+5] for i in range(0,len(text),5)]
    columns=t.pop(0)
    df=pd.DataFrame(t)
    df.columns=columns
    return df

def get_description(html):
    doc=pq(html)
    text=doc('.quote-sub-section p').text()
    df=pd.DataFrame([text])
    df.columns=['Description']
    return df



from pyquery import PyQuery as pq
import re
import pandas as pd


def get_stats(html):
    # parser for statistics in yahoo finance
    doc=pq(html)
    tables=doc('#YDC-Col1 table').items() # parsing rule
    # formatting to the clean format
    t=[i.text().split('\n') for i in tables]
    t=[t[0]]+[[i[j] for i in t[1:6] for j in range(len(i))]]+[[i[j] for i in t[7:9] for j in range(len(i))]]
    assert (len(t)==3)
    t=[[i[j:j+2] for j in range(0,len(i),2)] for i in t]
    for i in t: # formatting
        for j in i:
            j[0]=re.compile(' [0-9]$').sub('',j[0])
    assert (len(t)==3)
    dfs=[pd.DataFrame(i) for i in t]
    return dfs

def get_statements(html):
    # Not done yet, needs to format and pd
    doc = pq(html)
    items = doc('#mrt-node-Col1-1-Financials table tr').items()  # Parser, will be seperated
    result = [i.text().split('\n') for i in items if len(i.text().split('\n')) > 1]
    df = pd.DataFrame(result)
    return df
    
def get_reports(html): # parser for income, cash_flow, and balance sheets
    doc = pq(html)
    items = doc('#mrt-node-Col1-1-Financials table tr').items()  # Parsing rule
    result = [i.text().split('\n') for i in items if len(i.text().split('\n')) > 1] # formatting
    columns= result.pop(0)
    columns[0]='Statements' # eliminate unnecessary info
    df = pd.DataFrame(result)
    df.columns = columns
    return df





import os as _os


#__author__=='Hanbo'
#__version__=='0.0.1'

#global datapath
datapath=_os.path.join(_os.path.dirname(__file__), 'database')


from .stock import financials, analysis, profile, prices
from .stock.profile import description as desc

from .basics.stocks import stock_list

from .factors.factors import get_factors

from .operations.Save import save # done
from .operations.Tools import database_count, factors_names, sheets_names, code_count
from .operations.Open import open_file

#from operations.Open import open_file, open_general

#from operations.Trans import translate



# calculation.key.Beta is not done

from .calculation import *
from .get import update, getLastUpdate, main_get # Connected well

from .operations.Path import path # not tested

if len(_os.listdir(datapath))-2<100:
    print("There is not sufficient data in the database. Use main_get() to retrieve data")



class test:
    # Not supporting deliberate handlers of errors
    def __init__(self,stock):
        self.stock=stock
    def test_financials(self):
        df=financials.stats(self.stock)
        df=financials.Valuation_Measures(self.stock)
        df=financials.Financial_Highlights(self.stock)
        df=financials.Trading_Information(self.stock)
        df=financials.Income(self.stock)
        df=financials.Balance(self.stock)
        df=financials.Cash_Flow(self.stock)
    def test_analysis(self):
        df=analysis.EPS_Revisions(self.stock)
        df=analysis.EPS_Trend(self.stock)
        df=analysis.Earnings_Estimate(self.stock)
        df=analysis.Earnings_History(self.stock)
        df=analysis.Growth_Estimates(self.stock)
        df=analysis.Revenue_Estimate(self.stock)
    def test_profile(self):
        df=profile.Key_Executives(self.stock)
        df=profile.description(self.stock)
        df=profile.description(self.stock, 'zh')
        df=profile.summary(self.stock)
    def test_trading(self):
        df=prices.daily('BABA')
        df=prices.daily('BABA','2017-01-01')
        df=prices.daily('BABA','2017-01-01','2018-09-09')
        df=prices.weekly('BABA')
        df=prices.weekly('BABA','2017-01-01')
        df=prices.weekly('BABA','2017-01-01','2018-09-09')
        df=prices.monthly('BABA')
        df=prices.monthly('BABA','2017-01-01')
        df=prices.monthly('BABA','2017-01-01','2018-09-09')
    def test_stocks_list(self):
        df=stock_list()
        assert len(list(df))==9
        df=stock_list('NYSE')
        assert len(list(df))==9
        df=stock_list('NASDAQ')
        assert len(list(df))==9
    def test_calculation(self):
        pass
    



import os as _os
#import pandas as pd
import numpy as _np
from scipy import stats as _Stats

import gc as _gc


from ..operations.Format import factor as _factor
from ..operations.Path import path as _path
from ..operations.Open import open_file as _open_file
#from factors import _balance, cash_flow, key, stats

global _factor_dic
_factor_dic={'FCF':'cash_flow','IC':'balance','NIBCLS':'balance','Invested_Book_Capital':'balance'}


def factor_percentile(Factor,stock):
    '''
    Factor - str
    stock - company you want to check
    returns the percentile of the factor of the stock - a list of floats
    '''
    if not isinstance(Factor,str):
        raise TypeError("Parameters 'factors' should a string")
    p0=_os.path.abspath(_os.path.join(_os.path.dirname(__file__),_os.pardir))
    p=_os.path.join(p0,'database')
    name=_path(Factor)
    dirs=_os.listdir(p)
    d=[i for i in dirs if _os.path.isdir(_os.path.join(p,i))]
    #return d
    # i+'_'+name in set(os.listdir(os.path.join(path,i))
    r=[]
    for i in d:
        try:
            df=_open_file(i, name)
            f=_factor(df, Factor)
            if len(f)!=4:
                continue
            r.append(f)
        except Exception as e:
            #input('Exception: '+str(e))
            continue
    n = _np.stack(r)
    stock_value = _factor(_open_file(stock, name), Factor)
    result=[round(_Stats.percentileofscore(n[:,i],stock_value[i]),4) for i in range(len(stock_value))]
    del n,r,d
    _gc.collect()
    return result
        
def percentile(Factor, percentage=80):
    '''
    Factor - str
    percentile - int, percentage you want to check
    returns a list of companies with top percentile for the factor
    '''
    if not isinstance(Factor,str):
        raise TypeError("Parameters 'factors' should a string")
    p=_os.path.join(_os.path.abspath(_os.path.join(_os.path.dirname(__file__),_os.pardir)),'database')
    flag=False # flag for using external factors
    try:
        name=_path(Factor)
    except ValueError:
        if Factor in _factor_dic.keys():
            flag=True
            name=_factor_dic[Factor]
            
        else:
            raise ValueError("No support for factor '"+Factor+"'")
    dirs=_os.listdir(p)
    d=[i for i in dirs if _os.path.isdir(_os.path.join(p,i))]
    r=[]
    d2=d[:]
    for i in d2:
        try:
            df=_open_file(i,name)
            if flag:
                f=eval(name+'._'+Factor+'(df)') # call the function
            else:
                f=_factor(df,Factor)
            if len(f)!=4:
                d.remove(i)
                continue
            else:
                r.append(f)
                #input("Appended")
        except Exception as e:
            d.remove(i)
            #print(r)
            #input('Exception: '+str(e))
    n = _np.stack(r)
    assert len(n)==len(d)
    target = [_np.percentile(n[:,i], percentage) for i in range(4)]
    result=[[d[i] for i in range(len(n)) if n[i][j]>=target[j]] for j in range(len(target))]
    return result

def _CAGR(Factor, years): # compound anual growth rate, not developed
    pass



from pyquery import PyQuery as pq
import pandas as pd

def get_summary(html,stock):
    doc=pq(html)
    tables=doc('#quote-summary table').items()
    t=[i.text().split('\n') for i in tables]
    t=[i[j] for i in t for j in range(len(i))]
    columns=[t[i] for i in range(0,len(t),2)]
    data=[[t[i] for i in range(1,len(t),2)]]
    data[0].insert(0,stock)
    columns.insert(0,'Stock')
    del t
    df=pd.DataFrame(data)
    df.columns=columns
    return df

def get_news(html):
    doc=pq(html)
    pass



from .calculations import rank

#from operations.Open import open_file as _open_file

from .factors import cash_flow, income, balance, key, stats


def _factor(df, factor):
    f=_decimal(df[df['Statements'].isin([factor])].values[0][1:])
    return f






import pandas as _pd
import os as _os
import numpy as _np

from collections import Iterable as _iter


from ..operations.Open import open_file as _open_file
from ..operations.Path import path as _path
from ..operations.Format import factor as _factor


def get_factors(companies, factors, year='NEWEST'):
    # not handling year now
    '''
    companies - list of companies
    factors - list of factors
    year - default NEWEST, list of years
    companies, factors, year - maximum two of them can be 
    '''
    if isinstance(companies, str):
        companies=[companies] # string - list
    elif not isinstance(companies, _iter):
        raise TypeError("Parameter 'companies' should be iterable, not "
                        +type(companies).__name__) # iterable -> for later listcamp
    if isinstance(factors, list) and isinstance(year, list):
        if len(year)!=1 and len(factors)!=1 and len(companies)!=1:
            raise ValueError("When you want sheet for multiple factors with multiple companies, only a specific year is supported")
    if isinstance(factors, str):
        factors=[factors] # string - list
    paths = [_path(f) for f in factors] # things below is for drop duplicates
    d = {}
    for i in range(len(paths)):  # Avoid crowded IO
        if d.get(paths[i], 'no') == 'no':
            d[paths[i]] = [factors[i]]
        else:
            d[paths[i]] += [factors[i]]
    #for c in companies:
 #       a=_np.concatenate(tuple([_factor(_open_file(c, k), v, False) for k, v in d.items()]),
 #                         axis = 1)

    
    a=_np.concatenate(tuple
                      (
                          [_np.concatenate
                           (tuple
                            (
                                [_factor
                                 (_open_file(c, k), v, year) for k, v in d.items()
                                 ]
                                ),
                            axis = 1
                            )
                           for c in companies]
                          )
                      )
                     
    df=_pd.DataFrame(a, index=companies)
    # problem here
    df.columns=factors # when only one factor is here, the columns name should be date rather than factors
    return df


        
        


''' for debugging
factors=['Net Income','Total Assets','Total Revenue','Gross Profit']
paths = [_path(f) for f in factors]
d={}
for i in range(len(paths)):  # Avoid crowded IO
    if d.get(paths[i], 'no') == 'no':
        d[paths[i]] = [factors[i]]
    else:
        d[paths[i]] += [factors[i]]
df=_open_file('AAPL','income')
'''



from pyquery import PyQuery as pq
import pandas as pd

def get_analysis(html):
    doc=pq(html)
    tables=doc('#Main table').items()
    t=[i.text().split('\n') for i in tables]
    t_df=[[tt[i:i+4] for i in range(0,len(tt),5)] for tt in t]
    columns=[i.pop(0) for i in t_df]
    dfs=[pd.DataFrame(i) for i in t_df]
    for i in range(len(dfs)):
        dfs[i].columns=columns[i]
    return dfs



import numpy as _np


from ..operations.Format import factor as _factor
from ..operations.Open import open_file as _open_file

def FCF(stock): # Cash flow
    '''
    FCF - free cash flow, indicating companies 'real' cash, an important valuation measure
    stock - company name (e.g AAPL for apple inc.)
    '''
    df=_open_file(stock,'cash_flow')
    OCF=_factor(df, 'Total Cash Flow From Operating Activities')
    CE=_factor(df, 'Capital Expenditures')
    
    return _np.subtract(OCF,CE)




class cash_flow():
    def __init__(self,stock):
        self.__stock=stock
        self.__df=_open_file(stock,'cash_flow')
    def OCF(self):
        return _factor(self.__df, 'Total Cash Flow From Operating Activities')
    def CE(self):
        return _factor(self.__df, 'Capital Expenditures')
    def Net_Income(self):
        return _factor(self.__df, 'Net Income')
    def Depreciation(self):
        return _factor(self.__df, 'Depreciation')
    def Net_Borrowings(self):
        return _factor(self.__df, 'Net Borrowings')



import pandas as _pd


from ..operations.Open import open_general as _open_general


def stock_list(exchange=True):
    '''
    exchange - str - default True (all), either NYSE or NASDAQ
    returns a csv sheet with all statistics in yahoo finance for the company
    '''
    if exchange not in ['NYSE', 'NASDAQ']:
        raise ValueError("Parameter 'exchange' should either NYSE or NASDAQ")
    if exchange:
        c1=_open_general('NASDAQ.csv')
        c2=_open_general('NYSE.csv')
        df=_pd.concat([c1,c2],ignore_index=True).drop('Unnamed: 9',axis=1) # drop duplicated column
    else:
        _csv=_open_general(exchange+'.csv')
        df=_csv.drop('Unnamed: 9',axis=1)
    return df





import pandas as pd
from pyquery import PyQuery as pq


def get_major_holders(html):
    """Getter for holders from yahoo finance"""
    doc=pq(html)
    tables=doc('#mrt-node-Col1-1-Holders table') # parsing rule
    t1=pq(tables.pop(0)).text().split('\n')
    t2=[[t1[i+1],t1[i]] for i in range(0,len(t1),2)] # formatting
    df=pd.DataFrame(t2) # convert to pd dataframe
    df.columns=['Category','Percentage']
    return df
    

def get_top_institutional_and_mutual_fund_holders(html):
    """Getter for other holders in yahoo finance"""
    doc=pq(html)
    tables=doc('#mrt-node-Col1-1-Holders table') # parsing rule
    # formatting
    t=[pq(i).text().split('\n') for i in tables[1:]]
    t2=[[t1[i:i+4] for i in range(0,len(t1),5)] for t1 in t]
    columns=[t.pop(0) for t in t2]
    df1,df2=pd.DataFrame(t2[0]),pd.DataFrame(t2[1]) # convert to pd dataframe
    df1.columns=columns[0]
    df2.columns=columns[1]
    return (df1,df2)
    
    
    



import os

# __author__=='Hanbo'
# __version__=='0.0.1'

global datapath
datapath=os.path.join(os.path.dirname(__file__), 'database')


from stock.trading import (daily, weekly, monthly)

from stock.analysis import (Earnings_Estimate,Revenue_Estimate,
                            Earnings_History,EPS_Trend,EPS_Revisions,
                            Growth_Estimates)

from stock.financials import (stats, Valuation_Measures,
                              Financial_Highlights, Trading_Information,
                              Cash_Flow, Balance, Income)

from stock.profile import Key_Executives, description, summary

from basic.stock_list import stocks

from operations.Open import open_file, open_general

from operations.Trans import translate
